@@ -27,7 +27,9 @@ require "lolcat/version"
 require 'paint'
 
 module Lol
-  STRIP_ANSI = Regexp.compile '\e\[[\d;]*[m|K]', nil
+  ANSI_ESCAPE = /((?:\e(?:[ -\/]+.|[\]PX^_][^\a\e]*|\[[0-?]*.|.))*)(.?)/m
+  INCOMPLETE_ESCAPE = /\e(?:[ -\/]*|[\]PX^_][^\a\e]*|\[[0-?]*)$/
+
   @paint_detected_mode = Paint.detect_mode
 
   def self.rainbow(freq, i)
@@ -39,18 +41,32 @@ module Lol
 
   def self.cat(fd, opts={})
     print "\e[?25l" if opts[:animate]
-    fd.each do |line|
-      opts[:os] += 1
-      println(line, opts)
+    while true do
+      buf = ''
+      begin
+        begin
+          buf += fd.sysread(4096)
+          invalid_encoding = !buf.dup.force_encoding(fd.external_encoding).valid_encoding?
+        end while invalid_encoding or buf.match(INCOMPLETE_ESCAPE)
+      rescue EOFError
+        break
+      end
+      buf.force_encoding(fd.external_encoding)
+      buf.lines.each do |line|
+        opts[:os] += 1
+        println(line, opts)
+      end
     end
     ensure
-    print "\e[?25h" if opts[:animate]
+    if STDOUT.tty? then
+        print "\e[m\e[?25h\e[?1;5;2004l"
+        # system("stty sane -istrip <&1");
+    end
   end
 
   def self.println(str, defaults={}, opts={})
     opts.merge!(defaults)
-    chomped = str.chomp!
-    str.gsub! STRIP_ANSI, '' if !str.nil? and ($stdout.tty? or opts[:force])
+    chomped = str.sub!(/\n$/, "")
     str.gsub! "\t", "        "
     opts[:animate] ? println_ani(str, opts) : println_plain(str, opts)
     puts if chomped
@@ -61,18 +77,24 @@ module Lol
   def self.println_plain(str, defaults={}, opts={})
     opts.merge!(defaults)
     set_mode(opts[:truecolor])
-    str.chomp.chars.each_with_index do |c,i|
-      code = rainbow(opts[:freq], opts[:os]+i/opts[:spread])
-      print Paint[c, *[ (:black if opts[:invert]), code ].compact ]
+    str.scan(ANSI_ESCAPE).each_with_index do |c,i|
+      color = rainbow(opts[:freq], opts[:os]+i/opts[:spread])
+      if opts[:invert] then
+        print c[0], Paint.color(nil, color), c[1], "\e[49m"
+      else
+        print c[0], Paint.color(color), c[1], "\e[39m"
+      end
     end
   end
 
   def self.println_ani(str, opts={})
     return if str.empty?
+    print "\e7"
     (1..opts[:duration]).each do |i|
-      print "\e[#{str.length}D"
+      print "\e8"
       opts[:os] += opts[:spread]
       println_plain(str, opts)
+      str.gsub!(/\e\[[0-?]*[@JKPX]/, "")
       sleep 1.0/opts[:speed]
     end
   end
